import collections
import statistics as stat
from datetime import datetime
from .Models import (
    CAUTIONHOURTYPE_SUAVE,
    CAUTIONHOURTYPE_INTERMEDIATE,
    CAUTIONHOURTYPE_AGGRESSIVE,
    CAUTIONHOURTYPE
)


class HourObject:
    def __init__(self, nh:list, ch:list, dyn_ch:dict = {}) -> None:    
        self._nh = nh
        self._ch = ch
        self._dyn_ch = dyn_ch

    @property
    def nh(self) -> list:
        return self._nh
    
    @property
    def ch(self) -> list:
        return self._ch

    @property
    def dyn_ch(self) -> dict:
        return self._dyn_ch


class Hoursselectionbase:
    def __init__(
            self,      
            absolute_top_price: float = 0,
            min_price: float = 0,
            cautionhour_type: float = CAUTIONHOURTYPE[CAUTIONHOURTYPE_SUAVE]
    ):
        self._prices = None
        self._prices_tomorrow = None
        self._non_hours = []
        self._caution_hours = []
        self._dynamic_caution_hours = {}
        self._absolute_top_price = self._set_absolute_top_price(absolute_top_price)
        self._min_price = min_price
        self._cautionhour_type = cautionhour_type
        self._validate()
    
    def _set_absolute_top_price(self, val) -> float:
        if val is None:
            return float("inf")
        if val <= 0:
            return float("inf")
        return float(val)

    def _validate(self):
        assert 0 < self._cautionhour_type <= 1
        assert len(self._caution_hours) == 0
        assert len(self._non_hours) == 0

    @property
    def non_hours(self):
        return self._non_hours

    @non_hours.setter
    def non_hours(self, val):
        self._non_hours = val

    @property
    def caution_hours(self):
        return self._caution_hours

    @caution_hours.setter
    def caution_hours(self, val):
        self._caution_hours = val

    @property
    def dynamic_caution_hours(self) -> dict:
        return self._dynamic_caution_hours

    @dynamic_caution_hours.setter
    def dynamic_caution_hours(self, val):
        self._dynamic_caution_hours = val

    @property
    def prices(self):
        return self._prices

    @prices.setter
    def prices(self, val):
        self._prices = val
        self.update()

    @property
    def prices_tomorrow(self) -> list:
        return self._prices_tomorrow

    @prices_tomorrow.setter
    def prices_tomorrow(self, val):
        self._prices_tomorrow = self._convert_none_list(val)
        self.update()

    @property
    def absolute_top_price(self) -> float:
        return self._absolute_top_price

    @property
    def min_price(self) -> float:
        return self._min_price

    def update(self, testhour:int = None):
        hours_today = self._update_internal(self.prices)
        hours_tomorrow = self._update_internal(self.prices_tomorrow)
        hour = datetime.now().hour if testhour is None else testhour

        self.non_hours = []
        self.caution_hours = []
        self.dynamic_caution_hours = {}

        self.non_hours.extend(h for h in hours_today.nh if h >= hour)
        self.caution_hours.extend(h for h in hours_today.ch if h >= hour)
        self.non_hours.extend(h for h in hours_tomorrow.nh if h < hour)
        self.caution_hours.extend(h for h in hours_tomorrow.ch if h < hour)

        for h in hours_today.dyn_ch:
            if h >= hour:
                self._dynamic_caution_hours[h] = hours_today.dyn_ch[h]

        for h in hours_tomorrow.dyn_ch:
            if h < hour:
                self._dynamic_caution_hours[h] = hours_tomorrow.dyn_ch[h]

    def _update_internal(self, prices) -> HourObject:
        ret = HourObject([], [], {})
        if prices is not None and len(prices) > 1:
            pricedict = self._create_dict(prices)
            normalized_pricedict = self._create_dict(self._normalize_prices(prices))
            
            """
            Curve is too flat if stdev is <= 0.05. 
            If so we don't do any specific non or caution-hours based on pricing.
            """
            if stat.stdev(prices) > 0.05:
                prices_ranked = self._rank_prices(pricedict, normalized_pricedict)
                ready_hours = self._determine_hours(prices_ranked, prices)
            else:
                ready_hours = ret
            if self._absolute_top_price is not None:
                ret = self._add_expensive_non_hours(pricedict, ready_hours)
            else: 
                ret = ready_hours
            if self._min_price > 0:
                ret = self._remove_cheap_hours(pricedict, ret)
        return ret
        
    def _normalize_prices(self, prices) -> list:
        min_price = min(prices)
        ret = []
        for p in prices:
            ret.append(p/min_price)
        return ret

    def _remove_cheap_hours(self, hourdict: dict, hours: HourObject) -> HourObject:
        lst = (h for h in hourdict if hourdict[h] < self._min_price)
        for h in lst:
            if h in hours.nh:
                hours.nh.remove(h)
            elif h in hours.ch:
                hours.ch.remove(h)
                hours.dyn_ch.pop(h)    
        return hours

    def _add_expensive_non_hours(self, hourdict: dict, readyhours:HourObject) -> HourObject:
        lst = (h for h in hourdict if hourdict[h] >= self._absolute_top_price)
        for h in lst:
            if h not in readyhours.nh:
                readyhours.nh.append(h)
                if h in readyhours.ch:
                    readyhours.ch.remove(h)
                if h in readyhours.dyn_ch.keys():
                    readyhours.dyn_ch.pop(h)
        readyhours.nh.sort()
        return readyhours

    def _rank_prices(self, hourdict: dict, normalized_hourdict: dict) -> dict:
        ret = {}
        _maxval = max(hourdict.values())
        _max_normalized = max(normalized_hourdict.values())
        peaqstdev = _maxval/abs(_max_normalized/stat.stdev(normalized_hourdict.values()))
        if peaqstdev < min(hourdict.values()):
            peaqstdev = peaqstdev + min(hourdict.values())
        for key in hourdict:
            if hourdict[key] > peaqstdev:
                _permax = round(hourdict[key] / _maxval, 2)
                ret[key] = {"val": hourdict[key], "permax": _permax}
        return self._discard_excessive_hours(ret)

    def _discard_excessive_hours(self, hours: dict):
        """There should always be at least four regular hours before absolute_top_price kicks in."""
        if len(hours) < 20:
            return hours
        while len(hours) >= 20:
            to_pop = dict(sorted(hours.items(), key=lambda item: item[1]['val']))    
            hours.pop(list(to_pop.keys())[0])
        return hours

    def _create_dict(self, input: list):
        ret = {}
        for idx, val in enumerate(input):
            ret[idx] = val
        try:
            assert len(ret) == 24
        except Exception:
            raise ValueError
        return ret

    def _determine_hours(self, price_list: dict, prices: list) -> HourObject:
        _nh = []
        _dyn_ch = {}
        _ch = []
        for p in price_list:
            _permax = round(abs(price_list[p]["permax"] - 1), 2)
            if self._cautionhour_type == CAUTIONHOURTYPE[CAUTIONHOURTYPE_SUAVE]:
                _permax += 0.15
            elif self._cautionhour_type == CAUTIONHOURTYPE[CAUTIONHOURTYPE_INTERMEDIATE]:
                _permax += 0.05
            if float(price_list[p]["permax"]) <= self._cautionhour_type:
                _ch.append(p)
                _dyn_ch[p] = round(_permax,2)
            elif float(price_list[p]["val"]) <= (sum(prices)/len(prices)):
                _ch.append(p)
                _dyn_ch[p] = round(_permax,2)
            else:
                _nh.append(p)
        return HourObject(_nh, _ch, _dyn_ch)
    
    def _convert_none_list(self, lst:list) -> list:
        try:
            ret = []
            for l in lst:
                if l is None:
                    return ret
            return lst
        except:
            return self._make_array_from_empty(lst)

    def _make_array_from_empty(self, input) -> list:
        array = input.split(",")
        list = [p for p in array if len(p) > 0]
        ret = []
        if len(list) > 24:
            try:
                for l in list:
                    parsed_item = Hoursselectionbase._try_parse(l, float)
                    if not parsed_item:
                        parsed_item = Hoursselectionbase._try_parse(l, int)
                    assert type(parsed_item) is float or type(parsed_item) is int
                    ret.append(parsed_item)
                return ret
            except:
                return []
        return []

    def get_average_kwh_price(self, testhour:int = None):
        hour = datetime.now().hour if testhour is None else testhour
        ret = {}

        def _looper(h:int):
            if h in self._dynamic_caution_hours:
                    if self._prices_tomorrow is not None and len(self._prices_tomorrow) > 0:
                        if h < hour and len(self._prices_tomorrow) > 0:
                            ret[h] = self._dynamic_caution_hours[h] * self._prices_tomorrow[h]
                    if h >= hour:
                        ret[h] = self._dynamic_caution_hours[h] * self._prices[h]
            elif h not in self._non_hours:
                if h < hour and len(self._prices_tomorrow) > 0:
                    ret[h] = self._prices_tomorrow[h]
                if h >= hour:
                    ret[h] = self._prices[h]

        if self.prices_tomorrow is None:
            for h in range(hour,24):
                _looper(h)
        else:
            for h in range(hour,(hour+24)):
                h = h-24 if h > 23 else h
                _looper(h)
        
        return round(sum(ret.values())/len(ret),2)
        
    def get_total_charge(self, currentpeak:float, testhour:int = None) -> float:
        hour = datetime.now().hour if testhour is None else testhour
        ret = {}

        def _looper(h:int):
            if h in self._dynamic_caution_hours:
                    ret[h] = self._dynamic_caution_hours[h] * currentpeak
            elif h in self._non_hours:
                ret[h] = 0
            else:
                ret[h] = currentpeak

        if self.prices_tomorrow is None:
            for h in range(hour,24):
                _looper(h)
        else:
            for h in range(hour,(hour+24)):
                h = h-24 if h > 23 else h
                _looper(h)
        
        return round(sum(ret.values()),1)

    @staticmethod
    def _try_parse(input:str, parsetype:type):
        try:
            ret = parsetype(input)
            return ret
        except:
            return False



from .querytypes import (
    QUERYTYPE_AVERAGEOFTHREEDAYS,
    QUERYTYPE_AVERAGEOFTHREEDAYS_MIN,
    QUERYTYPE_AVERAGEOFTHREEHOURS_MIN, QUERYTYPE_AVERAGEOFTHREEHOURS,
    QUERYTYPE_BASICMAX,
    QUERYTYPES
)
from .locale_model import Locale_Type
from dataclasses import dataclass

@dataclass(frozen=True)
class NO_Tensio(Locale_Type):
    observed_peak = QUERYTYPE_AVERAGEOFTHREEDAYS_MIN
    charged_peak = QUERYTYPE_AVERAGEOFTHREEDAYS
    converted = True
    query_model = QUERYTYPES[QUERYTYPE_AVERAGEOFTHREEDAYS]

@dataclass(frozen=True)
class NO_LNett(Locale_Type):
    observed_peak = QUERYTYPE_AVERAGEOFTHREEHOURS_MIN
    charged_peak = QUERYTYPE_AVERAGEOFTHREEHOURS
    converted = True
    query_model = QUERYTYPES[QUERYTYPE_AVERAGEOFTHREEHOURS]

#docs: https://www.l-nett.no/nynettleie/slik-blir-ny-nettleie-og-pris        


@dataclass(frozen=True)
class NO_GlitreEnergi(Locale_Type):
    observed_peak = QUERYTYPE_BASICMAX
    charged_peak = QUERYTYPE_BASICMAX
    converted = True
    query_model = QUERYTYPES[QUERYTYPE_BASICMAX]

#docs: https://www.glitreenergi-nett.no/smart-nettleie/


@dataclass(frozen=True)
class NO_AgderEnergi(Locale_Type):
    observed_peak = QUERYTYPE_AVERAGEOFTHREEDAYS_MIN
    charged_peak = QUERYTYPE_AVERAGEOFTHREEDAYS
    converted = True    
    query_model = QUERYTYPES[QUERYTYPE_AVERAGEOFTHREEDAYS]



# import logging
# import time
# from datetime import datetime

# from custom_components.peaqev.peaqservice.util.chargerstates import CHARGECONTROLLER
# from custom_components.peaqev.peaqservice.util.constants import (
#     DOMAIN,
#     ON,
#     OFF,
#     RESUME,
#     PAUSE,
#     CHARGER,
#     PARAMS,
#     UPDATECURRENT,
#     CURRENT,
#     CHARGERID
# )

# _LOGGER = logging.getLogger(__name__)


# class Charger:
#     def __init__(self, hub, hass, servicecalls):
#         self._hass = hass
#         self._hub = hub
#         self._charger_running = False
#         self._charger_stopped = False
#         self._service_calls = servicecalls
#         self._session_is_active = False

#     async def charge(self):
#         """Main function to turn charging on or off"""
#         if self._hub.charger_enabled.value is True:
#             if self._hub.chargecontroller.status is CHARGECONTROLLER.Start:
#                 if self._hub.chargerobject_switch.value is False and self._charger_running is False:
#                     await self._start_charger()
#             elif self._hub.chargecontroller.status is CHARGECONTROLLER.Stop or self._hub.chargecontroller.status is CHARGECONTROLLER.Idle:
#                 if self._hub.chargerobject_switch.value is True and self._charger_stopped is False:
#                     await self._pause_charger()
#             elif self._hub.chargecontroller.status is CHARGECONTROLLER.Done and self._hub.charger_done.value is False:
#                 await self._terminate_charger()
#         else:
#             if self._hub.chargerobject_switch.value is True and self._charger_stopped is False:
#                 await self._terminate_charger()

#     async def _start_charger(self):
#         self._is_running(True)
#         if self._session_is_active is False:
#             await self._call_charger(ON)
#             self._session_is_active = True
#         else:
#             await self._call_charger(RESUME)
#         self._hub.chargecontroller.update_latestchargerstart()
#         if self._hub.chargertype.charger.servicecalls.allowupdatecurrent is True:
#             self._hass.async_create_task(self._updatemaxcurrent())

#     async def _terminate_charger(self):
#         self._is_running(False)
#         self._session_is_active = False
#         await self._call_charger(OFF)
#         self._hub.charger_done.value = True

#     async def _pause_charger(self):
#         self._is_running(False)
#         if self._hub.charger_done.value is True or self._hub.chargecontroller.status is CHARGECONTROLLER.Idle:
#             await self._terminate_charger()
#         else:
#             await self._call_charger(PAUSE)

#     async def _call_charger(self, command: str):
#         calls = self._service_calls.get_call(command)
#         await self._hub.hass.services.async_call(
#             calls[DOMAIN],
#             calls[command]
#         )

#     async def _updatemaxcurrent(self):
#         """If enabled, let the charger periodically update it's current during charging."""
#         calls = self._service_calls.get_call(UPDATECURRENT)
#         if await self._hass.async_add_executor_job(self._wait_turn_on):
#             while self._hub.chargerobject_switch.value is True and self._charger_running is True:
#                 if await self._hass.async_add_executor_job(self._wait_update_current):
#                     serviceparams = await self._setchargerparams(calls)
#                     info = f"peaqev updating current with: {calls[DOMAIN]}, {calls[UPDATECURRENT]} and params: {serviceparams}"
#                     _LOGGER.info(info)
#                     await self._hub.hass.services.async_call(
#                         calls[DOMAIN],
#                         calls[UPDATECURRENT],
#                         serviceparams
#                     )
#                     await self._hass.async_add_executor_job(self._wait_loop_cycle)

#             final_service_params = await self._setchargerparams(calls, ampoverride=6)
#             await self._hub.hass.services.async_call(
#                 calls[DOMAIN],
#                 calls[UPDATECURRENT],
#                 final_service_params
#             )

#     async def _setchargerparams(self, calls, ampoverride:int = 0) -> dict:
#         amps = ampoverride if ampoverride >= 6 else self._hub.threshold.allowedcurrent
#         serviceparams = {}
#         if await self._checkchargerparams(calls) is True:
#             serviceparams[calls[PARAMS][CHARGER]] = calls[PARAMS][CHARGERID]
#         serviceparams[calls[PARAMS][CURRENT]] = amps
#         return serviceparams

#     async def _checkchargerparams(self, calls) -> bool:
#         return len(calls[PARAMS][CHARGER]) > 0 and len(calls[PARAMS][CHARGERID]) > 0

#     def _wait_turn_on(self) -> bool:
#         while self._hub.chargerobject_switch.value is False and self._charger_stopped is False:
#             time.sleep(3)
#         if self._charger_stopped is True:
#             return False
#         return True

#     def _wait_update_current(self) -> bool:
#         while (self._hub.chargerobject_switch.current == self._hub.threshold.allowedcurrent
#                or (datetime.now().minute >= 55
#                    and self._hub.threshold.allowedcurrent > self._hub.chargerobject_switch.current)) \
#                 and self._charger_stopped is False:
#             time.sleep(3)
#         if self._charger_stopped is True:
#             return False
#         return True

#     def _wait_loop_cycle(self):
#         timer = 180
#         start_time = time.time()
#         while time.time() - start_time < timer:
#             time.sleep(3)

#     def _is_running(self, determinator: bool):
#         if determinator:
#             self._charger_running = True
#             self._charger_stopped = False
#         elif not determinator:
#             self._charger_running = False
#             self._charger_stopped = True



from datetime import datetime
from enum import Enum
from dataclasses import dataclass

"""Peak querytypes"""
QUERYTYPE_BASICMAX = "BasicMax"
QUERYTYPE_AVERAGEOFTHREEDAYS = "AverageOfThreeDays"
QUERYTYPE_AVERAGEOFTHREEHOURS = "AverageOfThreeHours"
QUERYTYPE_AVERAGEOFTHREEDAYS_MIN = "AverageOfThreeDays_Min"
QUERYTYPE_AVERAGEOFTHREEHOURS_MIN = "AverageOfThreeHours_Min"
QUERYTYPE_AVERAGEOFFIVEDAYS = "AverageOfFiveDays"
QUERYTYPE_AVERAGEOFFIVEDAYS_MIN = "AverageOfFiveDays_Min"
QUERYTYPE_HIGHLOAD = "HighLoad"
QUERYTYPE_AVERAGEOFTHREEHOURS_MON_FRI_07_19 = "sala"
QUERYTYPE_AVERAGEOFTHREEHOURS_MON_FRI_07_19_MIN = "sala"
QUERYTYPE_MAX_NOV_MAR_MON_FRI_06_22 = "skövde"
QUERYTYPE_BASICMAX_MON_FRI_07_17_DEC_MAR_ELSE_REGULAR = "kristinehamn"
QUERYTYPE_SOLLENTUNA = "sollentuna"
QUERYTYPE_SOLLENTUNA_MIN = "sollentuna_min"

"""Misc"""
QUARTER_HOURLY = "quarter-hourly"
HOURLY = "hourly"

class SumTypes(Enum):
    Max = 1
    Avg = 2
    Min = 3


class TimePeriods(Enum):
    Hourly = 1
    Daily = 2
    Weekly = 3
    BiWeekly = 4
    Monthly = 5
    Yearly = 6
    UnSet = 7


@dataclass(frozen=True)
class SumCounter:
    counter:int = 1
    groupby: TimePeriods = TimePeriods.UnSet


@dataclass(frozen=True)
class QueryProperties:
    sumtype: SumTypes
    timecalc:TimePeriods
    cycle: TimePeriods


@dataclass
class PeaksModel:
    p: dict
    m:int = 0

    def set_init_dict(self, dict_data):
        self.p = dict_data["p"]
        self.m = dict_data["m"]

class LocaleQuery:
    def __init__(
        self, 
        sumtype: SumTypes, 
        timecalc: TimePeriods, 
        cycle: TimePeriods, 
        sumcounter: SumCounter = None
        ) -> None:    
        self._peaks:PeaksModel = PeaksModel({})
        self._props = QueryProperties(
            sumtype, 
            timecalc, 
            cycle
            )
        self._sumcounter:SumCounter= sumcounter
        self._observed_peak_value:float = 0 
        self._charged_peak_value:float = 0

    @property
    def sumcounter(self) -> SumCounter:
        if self._sumcounter is not None:
            return self._sumcounter
        return SumCounter()

    @property
    def charged_peak(self) -> float: 
        return self._charged_peak_value

    @charged_peak.setter
    def charged_peak(self, val):
        self._charged_peak_value = val

    @property
    def observed_peak(self) -> float: 
        return self.charged_peak if self._props.sumtype is SumTypes.Max else self._observed_peak_value

    @observed_peak.setter
    def observed_peak(self, val):
        self._observed_peak_value = val

    def try_update(self, newval, dt = datetime.now()):
        _dt = (dt.day, dt.hour)
        if len(self._peaks.p) == 0:
            """first addition for this month"""
            self._peaks.p[_dt] = newval
            self._peaks.m = dt.month
        elif dt.month != self._peaks.m:
            """new month, reset"""
            self.reset_values(newval, dt)
        elif _dt in self._peaks.p.keys():
            self._set_update_for_groupby(newval, _dt)
        elif newval > min(self._peaks.p.values()):
            self._peaks.p[_dt] = newval
        elif len(self._peaks.p) < self.sumcounter.counter:
            self._peaks.p[_dt] = newval

        if len(self._peaks.p) > self.sumcounter.counter:
                self._peaks.p.pop(min(self._peaks.p, key=self._peaks.p.get))
        self._update_peaks()

    def _set_update_for_groupby(self, newval, _dt):
        if self.sumcounter.groupby in [TimePeriods.Daily, TimePeriods.UnSet]:
            _datekey = [k for k,v in self._peaks.p.items() if _dt[0] in k][0]
            if newval > self._peaks.p[_datekey]:
                    self._peaks.p.pop(_datekey)
                    self._peaks.p[_dt] = newval
        elif self.sumcounter.groupby is TimePeriods.Hourly:
            if newval > self._peaks.p[_dt]:
                    self._peaks.p[_dt] = newval

    def _update_peaks(self):
        if self._props.sumtype is SumTypes.Max:
            self.charged_peak = max(self._peaks.p.values())
        elif self._props.sumtype is SumTypes.Avg:
            self.observed_peak = min(self._peaks.p.values())
            self.charged_peak = sum(self._peaks.p.values()) / len(self._peaks.p)

    def reset_values(self, newval, dt):
        self._peaks.p.clear()
        self.try_update(newval, dt)

QUERYTYPES = {
    QUERYTYPE_AVERAGEOFTHREEHOURS: LocaleQuery(sumtype=SumTypes.Avg, timecalc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, sumcounter=SumCounter(counter=3, groupby=TimePeriods.Hourly)),
    QUERYTYPE_AVERAGEOFTHREEDAYS: LocaleQuery(sumtype=SumTypes.Avg, timecalc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, sumcounter=SumCounter(counter=3, groupby=TimePeriods.Daily)),
    QUERYTYPE_BASICMAX: LocaleQuery(sumtype=SumTypes.Max, timecalc=TimePeriods.Hourly, cycle=TimePeriods.Monthly)
}



from .Threshold import ThresholdBase as Threshold
from .Chargecontroller import ChargeControllerBase as ChargeController
from .Prediction import PredictionBase as Prediction
from .hoursselection import Hoursselectionbase as Hoursselection
from .session import SessionPrice as Session





import setuptools

setuptools.setup(
    name="peaqevcore",
    version="0.4.1",
    author="Magnus Eldén",
    description="Core types for peaqev car charging",
    url="https://github.com/elden1337/peaqev-core",
    license="MIT",
    packages=["peaqevcore"],
    test_requires=['pytest']
)   
